import ast
import sys
import textwrap
from unittest import mock

import pytest

from flake8_builtins import BuiltinsChecker


class FakeOptions:
    builtins_ignorelist = []
    builtins = None

    def __init__(self, ignore_list='', builtins=None):
        if ignore_list:
            self.builtins_ignorelist = ignore_list
        if builtins:
            self.builtins = builtins


def check_code(source, expected_codes=None, ignore_list=None, builtins=None):
    """Check if the given source code generates the given flake8 errors

    If `expected_codes` is a string is converted to a list,
    if it is not given, then it is expected to **not** generate any error.

    If `ignore_list` is provided, it should be a list of names
    that will be ignored if found, as if they were a builtin.

    If `builtins` is provided, it should be a list of names
    that will be reported if found, as if they were a builtin.
    """
    if isinstance(expected_codes, str):
        expected_codes = [expected_codes]
    elif expected_codes is None:
        expected_codes = []
    if ignore_list is None:
        ignore_list = []
    tree = ast.parse(textwrap.dedent(source))
    checker = BuiltinsChecker(tree, '/home/script.py')
    checker.parse_options(FakeOptions(ignore_list=ignore_list, builtins=builtins))
    return_statements = list(checker.run())

    assert len(return_statements) == len(expected_codes)

    for item, code in zip(return_statements, expected_codes):
        assert item[2].startswith(f'{code} ')


def test_builtin_top_level():
    source = 'max = 4'
    check_code(source, 'A001')


def test_ann_assign():
    source = 'list: int = 1'
    check_code(source, 'A001')


@pytest.mark.skipif(
    sys.version_info < (3, 8),
    reason='NamedExpr appeared in 3.8',
)
def test_walrus_operator():
    source = '(dict := 1)'
    check_code(source, 'A001')


def test_nested():
    source = """
    def bla():
        filter = 4
    """
    check_code(source, 'A001')


def test_more_nested():
    source = """
    class Bla(object):
        def method(self):
            int = 4
    """
    check_code(source, 'A001')


def test_line_number():
    source = """
    a = 2
    open = 4
    """
    tree = ast.parse(textwrap.dedent(source))
    checker = BuiltinsChecker(tree, '/home/script.py')
    checker.parse_options(FakeOptions())
    ret = list(checker.run())
    assert ret[0][0] == 3


def test_offset():
    source = """
    def bla():
        zip = 4
    """
    tree = ast.parse(textwrap.dedent(source))
    checker = BuiltinsChecker(tree, '/home/script.py')
    checker.parse_options(FakeOptions())
    ret = list(checker.run())
    assert ret[0][1] == 4


def test_assign_message():
    source = """
    def bla():
        object = 4
    """
    check_code(source, 'A001')


def test_assignment_starred():
    source = 'bla, *int = range(4)'
    check_code(source, 'A001')


def test_assignment_list():
    source = '[bla, int] = range(4)'
    check_code(source, 'A001')


def test_class_attribute_message():
    source = """
    class TestClass():
        object = 4
    """
    check_code(source, 'A003')


def test_argument_message():
    source = """
    def bla(list):
        a = 4"""
    check_code(source, 'A002')


def test_keyword_argument_message():
    source = """
    def bla(dict=3):
        b = 4"""
    check_code(source, 'A002')


def test_kwonly_argument_message():
    source = """
    def bla(*, list):
        a = 4
    """
    check_code(source, 'A002')


@pytest.mark.skipif(
    sys.version_info < (3, 8),
    reason='This syntax is only valid in Python 3.8+',
)
def test_posonly_argument_message():
    source = """
    def bla(list, /):
        a = 4
    """
    check_code(source, 'A002')


def test_no_error():
    source = """def bla(first):\n    b = 4"""
    check_code(source)


def test_method_without_arguments():
    source = """
    def bla():
        b = 4
    """
    check_code(source)


def test_method_only_normal_keyword_arguments():
    source = """
    def bla(k=4):
        b = 4
    """
    check_code(source)


def test_report_all_arguments():
    source = """
    def bla(zip, object=4):
        b = 4
    """
    check_code(source, ['A002', 'A002'])


def test_report_all_variables_within_a_line():
    source = """
    def bla():
        object = 4; zip = 3
    """
    check_code(source, ['A001', 'A001'])


def test_default_ignored_names():
    source = """
    class MyClass(object):
        __name__ = 4
    """
    check_code(source)


def test_custom_ignored_names():
    source = 'copyright = 4'
    check_code(source, ignore_list=('copyright',))


def test_flake8_builtins():
    source = 'consider_this_builtin = 4'
    check_code(source, ['A001'], builtins=('consider_this_builtin',))


def test_for_loop_variable():
    source = """
    for format in (1, 2, 3):
        continue
    """
    check_code(source, 'A001')


def test_for_loop_multiple_variables():
    source = """
    for (index, format) in enumerate([1,2,3,]):
        continue
    """
    check_code(source, 'A001')


def test_for_loop_list():
    source = """
    for [index, format] in enumerate([1,2,3,]):
        continue
    """
    check_code(source, 'A001')


def test_for_loop_nested_tuple():
    source = """
    for index, (format, list) in enumerate([(1, "a"), (2, "b")]):
        continue
    """
    check_code(source, ['A001', 'A001'])


def test_for_loop_starred():
    source = """
    for index, *int in enumerate([(1, "a"), (2, "b")]):
        continue
    """
    check_code(source, 'A001')


def test_for_loop_starred_no_error():
    source = """
    for index, *other in enumerate([(1, "a"), (2, "b")]):
        continue
    """
    check_code(source)


def test_with_statement():
    source = """
    with open("bla.txt") as dir:
        continue
    """
    check_code(source, 'A001')


def test_with_statement_no_error():
    source = 'with open("bla.txt"): ...'
    check_code(source)


def test_with_statement_multiple():
    source = 'with open("bla.txt") as dir, open("bla.txt") as int: ...'
    check_code(source, ['A001', 'A001'])


def test_with_statement_unpack():
    source = 'with open("bla.txt") as (dir, bla): ...'
    check_code(source, 'A001')


def test_with_statement_unpack_on_list():
    source = 'with open("bla.txt") as [dir, bla]: ...'
    check_code(source, 'A001')


def test_with_statement_unpack_star():
    source = 'with open("bla.txt") as (bla, *int): ...'
    check_code(source, 'A001')


def test_exception_py3():
    source = """
    try:
        a = 2
    except Exception as int: ...
    """
    check_code(source, 'A001')


def test_exception_no_error():
    source = """
    try:
        a = 2
    except Exception: ...
    """
    check_code(source)


def test_list_comprehension():
    source = 'a = [int for int in range(3,9)]'
    check_code(source, 'A001')


def test_set_comprehension():
    source = 'a = {int for int in range(3,9)}'
    check_code(source, 'A001')


def test_dict_comprehension():
    source = 'a = {int:"a" for int in range(3,9)}'
    check_code(source, 'A001')


def test_gen_comprehension():
    source = 'a = (int for int in range(3,9))'
    check_code(source, 'A001')


def test_list_comprehension_multiple():
    source = 'a = [(int, list) for int, list in enumerate(range(3,9))]\n'
    check_code(source, ['A001', 'A001'])


def test_list_comprehension_nested():
    source = 'a = [(int, str) for int in some() for str in other()]'
    check_code(source, ['A001', 'A001'])


def test_list_comprehension_multiple_as_list():
    source = 'a = [(int, a) for [int, a] in enumerate(range(3,9))]'
    check_code(source, 'A001')


def test_import_as():
    source = 'import zope.component.getSite as int'
    check_code(source, 'A001')


def test_import_from_as():
    source = 'from zope.component import getSite as int'
    check_code(source, 'A001')


def test_import_as_nothing():
    source = 'import zope.component.getSite as something_else'
    check_code(source)


def test_import_from_as_nothing():
    source = 'from zope.component import getSite as something_else'
    check_code(source)


def test_class():
    source = 'class int(object): ...'
    check_code(source, 'A001')


def test_class_nothing():
    source = 'class integer(object): ...'
    check_code(source)


def test_function():
    source = 'def int(): ...'
    check_code(source, 'A001')


def test_async_function():
    source = 'async def int(): ...'
    check_code(source, 'A001')


def test_method():
    source = """
    class bla(object):
        def int(): ...
    """
    check_code(source, 'A003')


def test_method_error_code():
    source = """
    class bla(object):
        def int(): ...
    """
    check_code(source, 'A003')


def test_function_nothing():
    source = 'def integer(): ...'
    check_code(source)


def test_async_for():
    source = """
    async def bla():
        async for int in range(4): ...
    """
    check_code(source, 'A001')


def test_async_for_nothing():
    source = """
    async def bla():
        async for x in range(4): ...
    """
    check_code(source)


def test_async_with():
    source = """
    async def bla():
        async with open("bla.txt") as int: ...
    """
    check_code(source, 'A001')


def test_async_with_nothing():
    source = """
    async def bla():
        async with open("bla.txt") as x: ...
    """
    check_code(source)


@mock.patch('flake8.utils.stdin_get_value')
def test_stdin(stdin_get_value):
    source = 'max = 4'
    stdin_get_value.return_value = source
    checker = BuiltinsChecker('', 'stdin')
    checker.parse_options(FakeOptions())
    ret = list(checker.run())
    assert len(ret) == 1


def test_tuple_unpacking():
    source = 'a, *(b, c) = 1, 2, 3'
    check_code(source)



import ast
import builtins
import inspect
import sys

from flake8 import utils as stdin_utils

if sys.version_info >= (3, 8):
    NamedExpr = ast.NamedExpr
else:  # There was no walrus operator before python3.8
    NamedExpr = type('NamedExpr', (ast.AST,), {})


class BuiltinsChecker:
    name = 'flake8_builtins'
    version = '1.5.2'
    assign_msg = 'A001 variable "{0}" is shadowing a python builtin'
    argument_msg = 'A002 argument "{0}" is shadowing a python builtin'
    class_attribute_msg = 'A003 class attribute "{0}" is shadowing a python builtin'

    names = []
    ignore_list = {
        '__name__',
        '__doc__',
        'credits',
        '_',
    }

    def __init__(self, tree, filename):
        self.tree = tree
        self.filename = filename

    @classmethod
    def add_options(cls, option_manager):
        option_manager.add_option(
            '--builtins-ignorelist',
            metavar='builtins',
            parse_from_config=True,
            comma_separated_list=True,
            help='A comma separated list of builtins to skip checking',
        )

    @classmethod
    def parse_options(cls, options):
        if options.builtins_ignorelist is not None:
            cls.ignore_list.update(options.builtins_ignorelist)

        cls.names = {
            a[0] for a in inspect.getmembers(builtins) if a[0] not in cls.ignore_list
        }
        flake8_builtins = getattr(options, 'builtins', None)
        if flake8_builtins:
            cls.names.update(flake8_builtins)

    def run(self):
        tree = self.tree

        if self.filename == 'stdin':
            lines = stdin_utils.stdin_get_value()
            tree = ast.parse(lines)

        for statement in ast.walk(tree):
            for child in ast.iter_child_nodes(statement):
                child.__flake8_builtins_parent = statement

        function_nodes = [ast.FunctionDef]
        if getattr(ast, 'AsyncFunctionDef', None):
            function_nodes.append(ast.AsyncFunctionDef)
        function_nodes = tuple(function_nodes)

        for_nodes = [ast.For]
        if getattr(ast, 'AsyncFor', None):
            for_nodes.append(ast.AsyncFor)
        for_nodes = tuple(for_nodes)

        with_nodes = [ast.With]
        if getattr(ast, 'AsyncWith', None):
            with_nodes.append(ast.AsyncWith)
        with_nodes = tuple(with_nodes)

        comprehension_nodes = (
            ast.ListComp,
            ast.SetComp,
            ast.DictComp,
            ast.GeneratorExp,
        )

        value = None
        for statement in ast.walk(tree):
            if isinstance(statement, (ast.Assign, ast.AnnAssign, NamedExpr)):
                value = self.check_assignment(statement)

            elif isinstance(statement, function_nodes):
                value = self.check_function_definition(statement)

            elif isinstance(statement, for_nodes):
                value = self.check_for_loop(statement)

            elif isinstance(statement, with_nodes):
                value = self.check_with(statement)

            elif isinstance(statement, ast.excepthandler):
                value = self.check_exception(statement)

            elif isinstance(statement, comprehension_nodes):
                value = self.check_comprehension(statement)

            elif isinstance(statement, (ast.Import, ast.ImportFrom)):
                value = self.check_import(statement)

            elif isinstance(statement, ast.ClassDef):
                value = self.check_class(statement)

            if value:
                yield from value

    def check_assignment(self, statement):
        msg = self.assign_msg
        if type(statement.__flake8_builtins_parent) is ast.ClassDef:
            msg = self.class_attribute_msg

        if isinstance(statement, ast.Assign):
            stack = list(statement.targets)
        else:  # This is `ast.AnnAssign` or `ast.NamedExpr`:
            stack = [statement.target]

        while stack:
            item = stack.pop()
            if isinstance(item, (ast.Tuple, ast.List)):
                stack.extend(list(item.elts))
            elif isinstance(item, ast.Name) and item.id in self.names:
                yield self.error(item, message=msg, variable=item.id)
            elif isinstance(item, ast.Starred):
                if hasattr(item.value, 'id') and item.value.id in self.names:
                    yield self.error(
                        statement,
                        message=msg,
                        variable=item.value.id,
                    )
                elif hasattr(item.value, 'elts'):
                    stack.extend(list(item.value.elts))

    def check_function_definition(self, statement):
        if statement.name in self.names:
            msg = self.assign_msg
            if type(statement.__flake8_builtins_parent) is ast.ClassDef:
                msg = self.class_attribute_msg

            yield self.error(statement, message=msg, variable=statement.name)

        all_arguments = []
        all_arguments.extend(statement.args.args)
        all_arguments.extend(getattr(statement.args, 'kwonlyargs', []))
        all_arguments.extend(getattr(statement.args, 'posonlyargs', []))

        for arg in all_arguments:
            if isinstance(arg, ast.arg) and arg.arg in self.names:
                yield self.error(
                    arg,
                    message=self.argument_msg,
                    variable=arg.arg,
                )

    def check_for_loop(self, statement):
        stack = [statement.target]
        while stack:
            item = stack.pop()
            if isinstance(item, (ast.Tuple, ast.List)):
                stack.extend(list(item.elts))
            elif isinstance(item, ast.Name) and item.id in self.names:
                yield self.error(statement, variable=item.id)
            elif isinstance(item, ast.Starred):
                if hasattr(item.value, 'id') and item.value.id in self.names:
                    yield self.error(
                        statement,
                        variable=item.value.id,
                    )
                elif hasattr(item.value, 'elts'):
                    stack.extend(list(item.value.elts))

    def check_with(self, statement):
        for item in statement.items:
            var = item.optional_vars
            if isinstance(var, (ast.Tuple, ast.List)):
                for element in var.elts:
                    if isinstance(element, ast.Name) and element.id in self.names:
                        yield self.error(statement, variable=element.id)
                    elif (
                        isinstance(element, ast.Starred)
                        and element.value.id in self.names
                    ):
                        yield self.error(
                            element,
                            variable=element.value.id,
                        )

            elif isinstance(var, ast.Name) and var.id in self.names:
                yield self.error(statement, variable=var.id)

    def check_exception(self, statement):
        exception_name = statement.name
        if exception_name is None:
            return

        if exception_name in self.names:
            yield self.error(statement, variable=exception_name)

    def check_comprehension(self, statement):
        for generator in statement.generators:
            if (
                isinstance(generator.target, ast.Name)
                and generator.target.id in self.names
            ):
                yield self.error(statement, variable=generator.target.id)

            elif isinstance(generator.target, (ast.Tuple, ast.List)):
                for tuple_element in generator.target.elts:
                    if (
                        isinstance(tuple_element, ast.Name)
                        and tuple_element.id in self.names
                    ):
                        yield self.error(statement, variable=tuple_element.id)

    def check_import(self, statement):
        for name in statement.names:
            if name.asname in self.names:
                yield self.error(statement, variable=name.asname)

    def check_class(self, statement):
        if statement.name in self.names:
            yield self.error(statement, variable=statement.name)

    def error(self, statement, variable, message=None):
        if not message:
            message = self.assign_msg

        return (
            statement.lineno,
            statement.col_offset,
            message.format(variable),
            type(self),
        )
